"""
PuePy's router functionality can be optionally installed by calling the install_router method of the Application class.

Example:
    ``` py
    from puepy import Application, Router

    app = Application()
    app.install_router(Router, link_mode=Router.LINK_MODE_HASH)
    ```

Once installed, the `Router` instance is available on `app.Router` and can be used throughout the application to manage
client-side routing. Routes are defined by either using the `@app.page` decorator or by calling methods manually on
the `Router` instance.

Classes:
    puepy.router.Route: Represents a route in the router.
    puepy.router.Router: Represents a router for managing client-side routing in a web application.
"""

from .core import Page
from .runtime import window, history, platform, PLATFORM_MICROPYTHON, is_server_side
from .util import mixed_to_underscores, jsobj


def _micropython_parse_query_string(query_string):
    """
    In MicroPython, urllib isn't available and we can't use the JavaScript library:
    https://github.com/pyscript/pyscript/issues/2100
    """
    if query_string and query_string[0] == "?":
        query_string = query_string[1:]

    def url_decode(s):
        # Decode URL-encoded characters without using regex, which is also pretty broken in MicroPython...
        i = 0
        length = len(s)
        decoded = []

        while i < length:
            if s[i] == "%":
                if i + 2 < length:
                    hex_value = s[i + 1 : i + 3]
                    decoded.append(chr(int(hex_value, 16)))
                    i += 3
                else:
                    decoded.append("%")
                    i += 1
            elif s[i] == "+":
                decoded.append(" ")
                i += 1
            else:
                decoded.append(s[i])
                i += 1

        return "".join(decoded)

    params = {}
    for part in query_string.split("&"):
        if "=" in part:
            key, value = part.split("=", 1)
            key = url_decode(key)
            value = url_decode(value)
            if key in params:
                params[key].append(value)
            else:
                params[key] = [value]
        else:
            key = url_decode(part)
            if key in params:
                params[key].append("")
            else:
                params[key] = ""
    return params


if platform == PLATFORM_MICROPYTHON:
    from js import encodeURIComponent

    def url_quote(s):
        return encodeURIComponent(s)

else:
    from urllib.parse import quote as url_quote

if is_server_side:
    from urllib.parse import urlparse, parse_qs

    def parse_query_string(qs):
        return parse_qs(urlparse(qs).query)

elif platform == PLATFORM_MICROPYTHON:
    parse_query_string = _micropython_parse_query_string

else:
    import js

    def parse_query_string(qs):
        usp = js.URLSearchParams.new(qs)
        return {key: list(usp.getAll(key)) for key in usp.keys()}


class Route:
    """
    Represents a route in the router. A route is defined by a path match pattern, a page class, and a name.

    Note:
        This is usually not instanciated directly. Instead, use the `Router.add_route` method to create a new route or
        use the @app.page decorator to define a route at the time you define your Pages.
    """

    def __init__(self, path_match: str, page: Page, name: str, base_path: str, router=None):
        """
        Args:
            path_match (str): The path match pattern used for routing.
            page (Page): An instance of the Page class representing the page.
            name (str): The name of the page.
            base_path (str): The base path used for routing.
            router (Router, optional): An optional parameter representing the router used for routing.
        """
        self.path_match = path_match
        self.page = page
        self.name = name
        self.base_path = base_path
        self.router = router

    def match(self, path):
        """
        Evaluates a path against the route's pattern to determine if there is a match.

        Args:
            path: The path to be matched against the pattern.

        Returns:
            Match found (tuple): A tuple containing a True boolean value and a dictionary. The dictionary contains the
            matched variables extracted from the path.

            Match not found (tuple): If no match is found, returns `(False, None)`.
        """
        if self.base_path and path.startswith(self.base_path):
            path = path[len(self.base_path) :]

        # Simple pattern matching without regex
        parts = path.strip("/").split("/")
        pattern_parts = self.path_match.strip("/").split("/")
        if len(parts) != len(pattern_parts):
            return False, None

        kwargs = {}
        for part, pattern_part in zip(parts, pattern_parts):
            if pattern_part.startswith("<") and pattern_part.endswith(">"):
                group_name = pattern_part[1:-1]
                kwargs[group_name] = part
            elif part != pattern_part:
                return False, None

        return True, kwargs

    def reverse(self, **kwargs):
        """
        Reverse method is used to generate a URL path using the given parameters. It replaces the placeholders in the
        path template with the corresponding values.

        Args:
            **kwargs: A variable number of keyword arguments representing the values to be inserted into the path
            template.

        Returns:
            (str): The generated URL path.

        Example:
            Let's say we have a path template `/users/<username>/posts/<post_id>`. We can use the reverse method to
            generate the URL path by providing the values for "username" and "post_id" as keyword arguments:
            `route.reverse(username="john", post_id=123)` => `"/users/john/posts/123"`
        """
        kwargs = kwargs.copy()
        result = self.path_match
        for key in list(kwargs.keys()):
            if f"<{key}>" in result:
                value = kwargs.pop(key)
                result = result.replace(f"<{key}>", str(value))

        if self.router and self.router.link_mode == Router.LINK_MODE_HASH:
            result = "#" + result

        if self.base_path:
            path = f"{self.base_path}{result}"
        else:
            path = result

        if kwargs:
            path += "?" + "&".join(f"{url_quote(k)}={url_quote(v)}" for k, v in kwargs.items())
        return path

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"<Route: {self.name}>"


class Router:
    """Class representing a router for managing client-side routing in a web application.



    Args:
        application (object, optional): The web application object. Defaults to None.
        base_path (str, optional): The base path URL. Defaults to None.
        link_mode (str, optional): The link mode for navigating. Defaults to "hash".

    Attributes:
        LINK_MODE_DIRECT (str): Direct link mode.
        LINK_MODE_HTML5 (str): HTML5 link mode.
        LINK_MODE_HASH (str): Hash link mode.
        routes (list): List of Route instances.
        routes_by_name (dict): Dictionary mapping route names to Route instances.
        routes_by_page (dict): Dictionary mapping page classes to Route instances.
        application (object): The web application object.
        base_path (str): The base path URL.
        link_mode (str): The link mode for navigating.
    """

    LINK_MODE_DIRECT = "direct"
    LINK_MODE_HTML5 = "html5"
    LINK_MODE_HASH = "hash"

    def __init__(self, application=None, base_path=None, link_mode=LINK_MODE_HASH):
        """
        Initializes an instance of the class.

        Parameters:
            application (Application): The application used for routing.
            base_path (str): The base path for the routes.
            link_mode (str): The mode for generating links.
        """
        self.routes = []
        self.routes_by_name = {}
        self.routes_by_page = {}
        self.application = application
        self.base_path = base_path
        self.link_mode = link_mode

    def add_route_instance(self, route: Route):
        """
        Add a route instance to the current router.

        Parameters:
            route (Route): The route instance to be added.

        Raises:
            ValueError: If the route instance or route name already exists in the router.
        """
        if route in self.routes:
            raise ValueError(f"Route already added: {route}")
        if route.name in self.routes_by_name:
            raise ValueError(f"Route name already exists for another route: {route.name}")
        self.routes.append(route)
        self.routes_by_name[route.name] = route
        self.routes_by_page[route.page] = route
        route.router = self

    def add_route(self, path_match, page_class, name=None):
        """
        Adds a route to the router. This method creates a new Route instance.

        Args:
            path_match (str): The URL path pattern to match for the route.
            page_class (Page class): The class or function to be associated with the route.
            name (str, optional): The name of the route. If not provided, the name will be derived from the page class name.
        """
        # Convert path to a simple pattern without regex
        if not name:
            name = mixed_to_underscores(page_class.__name__)
        self.add_route_instance(Route(path_match=path_match, page=page_class, name=name, base_path=self.base_path))

    def reverse(self, destination, **kwargs):
        """
        Reverses a

        Args:
            destination: The destination to reverse. It can be the name of a route, the mapped page of a route, or the default page of the application.
            **kwargs: Additional keyword arguments to be passed to the reverse method of the destination route.

        Returns:
            (str): The reversed URL for the given destination.

        Raises:
            KeyError: If the destination is not found in the routes.
        """
        route: Route
        if isinstance(destination, Route):
            return destination.reverse(**kwargs)
        elif destination in self.routes_by_name:
            route = self.routes_by_name[destination]
        elif destination in self.routes_by_page:
            route = self.routes_by_page[destination]
        elif self.application and destination is self.application.default_page:
            if self.link_mode == Router.LINK_MODE_HASH:
                path = "#/"
            else:
                path = "/"
            return self.base_path or "" + path
        else:
            raise KeyError(f"{destination} not found in routes")
        return route.reverse(**kwargs)

    def match(self, path):
        """
        Args:
            path (str): The path to be matched.

        Returns:
            (tuple): A tuple containing the matching route and the matched route arguments (if any). If no route is
                found, returns (None, None).
        """
        path = path.split("#")[0]
        if "?" not in path:
            path += "?"
        path, query_string = path.split("?", 1)
        arguments = parse_query_string(query_string)

        for route in self.routes:
            matches, path_arguments = route.match(path)
            if path_arguments:
                arguments.update(path_arguments)
            if matches:
                return route, arguments
        return None, None

    def navigate_to_path(self, path, **kwargs):
        """
        Navigates to the specified path.

        Args:
            path (str or Page): The path to navigate to. If path is a subclass of Page, it will be reversed using the reverse method
            provided by the self object. If path is a string and **kwargs is not empty, it will append the query string
            to the path.

            **kwargs: Additional key-value pairs to be included in the query string. Each key-value pair will be
            URL-encoded.

        Raises:
            Exception: If the link mode is invalid.
        """
        if isinstance(path, type) and issubclass(path, Page):
            path = self.reverse(path, **kwargs)
        elif kwargs:
            path += "?" + "&".join(f"{url_quote(k)}={url_quote(v)}" for k, v in kwargs.items())

        if self.link_mode == self.LINK_MODE_DIRECT:
            window.location = path
        elif self.link_mode == self.LINK_MODE_HTML5:
            history.pushState(jsobj(), "", path)
            self.application.mount(self.application._selector_or_element, path)
        elif self.link_mode == self.LINK_MODE_HASH:
            path = path[1:] if path.startswith("#") else path
            if not is_server_side:
                history.pushState(jsobj(), "", "#" + path)
            self.application.mount(self.application._selector_or_element, path)
        else:
            raise Exception(f"Invalid link mode: {self.link_mode}")



import itertools
import sys

from puepy.storage import BrowserStorage

from . import exceptions
from .core import Page, t, Prop
from .reactivity import ReactiveDict, Stateful
from .runtime import (
    is_server_side,
    add_event_listener,
    window,
)


class GenericErrorPage(Page):
    props = ["error"]

    def populate(self):
        t.h1(f"Error: {self.error}")


class TracebackErrorPage(Page):
    props = ["error"]

    def format_exception(self):
        try:
            import traceback
        except ImportError:
            import io

            buf = io.StringIO()
            sys.print_exception(self.error, buf)
            return buf.getvalue()

        return "\n".join(traceback.format_exception(type(self.error), self.error, self.error.__traceback__))

    def populate(self):
        t.h1(f"Error: {self.error}")
        t.pre(str(self.format_exception()))


class DefaultIdGenerator:
    """
    A class representing a default ID generator.

    The DefaultIdGenerator class generates serial IDs with an optional prefix using a counter and base36 encoding.

    Attributes:
        counter (itertools.count): A counter object to keep track of the next ID value.
        prefix (str): An optional prefix to prepend to the generated IDs.
    """

    def __init__(self, prefix="pp-"):
        """
        Args:
            prefix (Optional[str]): The prefix to be used for generating unique IDs. Default is "pp-".
        """
        self.counter = itertools.count()
        self.prefix = prefix

    @staticmethod
    def _int_to_base36(num):
        base_chars = "0123456789abcdefghijklmnopqrstuvwxyz"
        base = len(base_chars)
        encoded = ""
        while num:
            num, rem = divmod(num, base)
            encoded = base_chars[rem] + encoded
        return encoded or base_chars[0]

    def get_id_for_element(self, element):
        """
        Returns the unique ID for the given element.

        Args:
            element: The element for which the ID is to be retrieved.

        Returns:
            The unique ID for the given element.
        """
        return self.prefix + self._int_to_base36(next(self.counter))


class Application(Stateful):
    """
    The main application class for PuePy. It manages the state, storage, router, and pages for the application.

    Attributes:
        state (ReactiveDict): The state object for the application.
        session_storage (BrowserStorage): The session storage object for the application.
        local_storage (BrowserStorage): The local storage object for the application.
        router (Router): The router object for the application, if any
        default_page (Page): The default page to mount if no route is matched.
        active_page (Page): The currently active page.
        not_found_page (Page): The page to mount when a 404 error occurs.
        forbidden_page (Page): The page to mount when a 403 error occurs.
        unauthorized_page (Page): The page to mount when a 401 error occurs.
        error_page (Page): The page to mount when an error occurs.
    """

    def __init__(self, element_id_generator=None):
        self.state = ReactiveDict(self.initial())
        self.add_context("state", self.state)

        if is_server_side:
            self.session_storage = None
            self.local_storage = None
        else:
            from js import localStorage, sessionStorage

            self.session_storage = BrowserStorage(sessionStorage, "session_storage")
            self.local_storage = BrowserStorage(localStorage, "local_storage")
        self.router = None
        self._selector_or_element = None
        self.default_page = None
        self.active_page = None

        self.not_found_page = GenericErrorPage
        self.forbidden_page = GenericErrorPage
        self.unauthorized_page = GenericErrorPage
        self.error_page = TracebackErrorPage

        self.element_id_generator = element_id_generator or DefaultIdGenerator()

    def install_router(self, router_class, **kwargs):
        """
        Install a router in the application.

        Args:
            router_class (class): A class that implements the router logic for the application. At this time, only
                `puepy.router.Router` is available.
            **kwargs: Additional keyword arguments that can be passed to the router_class constructor.
        """
        self.router = router_class(application=self, **kwargs)
        if not is_server_side:
            add_event_listener(window, "popstate", self._on_popstate)

    def page(self, route=None, name=None):
        """
        A decorator for `Page` classes which adds the page to the application with a specified route and name.

        Intended to be called as a decorator.

        Args:
            route (str): The route for the page. Default is None.
            name (str): The name of the page. If left None, page class is used as the name.

        Examples:
            ``` py
            app = Application()
            @app.page("/my-page")
            class MyPage(Page):
                ...
            ```
        """
        if route:
            if not self.router:
                raise Exception("Router not installed")

            def decorator(func):
                self.router.add_route(route, func, name=name)
                return func

            return decorator
        else:

            def decorator(func):
                self.default_page = func
                return func

            return decorator

    def _on_popstate(self, event):
        if self.router.link_mode == self.router.LINK_MODE_HASH:
            self.mount(self._selector_or_element, window.location.hash.split("#", 1)[-1])
        elif self.router.link_mode in (self.router.LINK_MODE_DIRECT, self.router.LINK_MODE_HTML5):
            self.mount(self._selector_or_element, window.location.pathname)

    def remount(self, path=None, page_kwargs=None):
        """
        Remounts the selected element or selector with the specified path and page_kwargs.

        Args:
            path (str): The new path to be used for remounting the element or selector. Default is None.
            page_kwargs (dict): Additional page kwargs to be passed when remounting. Default is None.

        """
        self.mount(self._selector_or_element, path=path, page_kwargs=page_kwargs)

    def mount(self, selector_or_element, path=None, page_kwargs=None):
        """
        Mounts a page onto the specified selector or element with optional path and page_kwargs.

        Args:
            selector_or_element: The selector or element on which to mount the page.
            path: Optional path to match against the router. Defaults to None.
            page_kwargs: Optional keyword arguments to pass to the mounted page. Defaults to None.

        Returns:
            (Page): The mounted page instance
        """
        if page_kwargs is None:
            page_kwargs = {}

        self._selector_or_element = selector_or_element

        if self.router:
            path = path or self.current_path
            route, arguments = self.router.match(path)
            if arguments:
                page_kwargs.update(arguments)

            if route:
                page_class = route.page
            elif path in ("", "/") and self.default_page:
                page_class = self.default_page
            elif self.not_found_page:
                page_class = self.not_found_page
            else:
                return None
        elif self.default_page:
            route = None
            page_class = self.default_page
        else:
            return None

        self.active_page = None
        try:
            self.mount_page(
                selector_or_element=selector_or_element,
                page_class=page_class,
                route=route,
                page_kwargs=page_kwargs,
                handle_exceptions=True,
            )
        except Exception as e:
            self.handle_error(e)
        return self.active_page

    @property
    def current_path(self):
        """
        Returns the current path based on the router's link mode.

        Returns:
            str: The current path.
        """
        if self.router.link_mode == self.router.LINK_MODE_HASH:
            return window.location.hash.split("#", 1)[-1]
        elif self.router.link_mode in (self.router.LINK_MODE_DIRECT, self.router.LINK_MODE_HTML5):
            return window.location.pathname
        else:
            return ""

    def mount_page(self, selector_or_element, page_class, route, page_kwargs, handle_exceptions=True):
        """
        Mounts a page on the specified selector or element with the given parameters.

        Args:
            selector_or_element (str or Element): The selector string or element to mount the page on.
            page_class (class): The page class to mount.
            route (str): The route for the page.
            page_kwargs (dict): Additional keyword arguments to pass to the page class.
            handle_exceptions (bool, optional): Determines whether to handle exceptions thrown during mounting.
                Defaults to True.
        """
        page_class._expanded_props()

        # For security, we only pass props to the page that are defined in the page's props
        #
        # We also handle the list or not-list props for multiple or single values
        # (eg, ?foo=1&foo=2 -> ["1", "2"] if needed)
        #
        prop_args = {}
        prop: Prop
        for prop in page_class.props_expanded.values():
            if prop.name in page_kwargs:
                value = page_kwargs.pop(prop.name)
                if prop.type is list:
                    prop_args[prop.name] = value if isinstance(value, list) else [value]
                else:
                    prop_args[prop.name] = value if not isinstance(value, list) else value[0]

        self.active_page: Page = page_class(matched_route=route, application=self, extra_args=page_kwargs, **prop_args)
        try:
            self.active_page.mount(selector_or_element)
        except exceptions.PageError as e:
            if handle_exceptions:
                self.handle_page_error(e)
            else:
                raise

    def handle_page_error(self, exc):
        """
        Handles page error based on the given exception by inspecting the exception type and passing it along to one
        of:

        - `handle_not_found`
        - `handle_forbidden`
        - `handle_unauthorized`
        - `handle_redirect`
        - `handle_error`

        Args:
            exc (Exception): The exception object representing the page error.
        """
        if isinstance(exc, exceptions.NotFound):
            self.handle_not_found(exc)
        elif isinstance(exc, exceptions.Forbidden):
            self.handle_forbidden(exc)
        elif isinstance(exc, exceptions.Unauthorized):
            self.handle_unauthorized(exc)
        elif isinstance(exc, exceptions.Redirect):
            self.handle_redirect(exc)
        else:
            self.handle_error(exc)

    def handle_not_found(self, exception):
        """
        Handles the exception for not found page. By default, it mounts the self.not_found_page class and passes it
        the exception as an argument.

        Args:
            exception (Exception): The exception that occurred.
        """
        self.mount_page(
            self._selector_or_element, self.not_found_page, None, {"error": exception}, handle_exceptions=False
        )

    def handle_forbidden(self, exception):
        """
        Handles the exception for forbidden page. By default, it mounts the self.forbidden_page class and passes it
        the exception as an argument.

        Args:
            exception (Exception): The exception that occurred.
        """
        self.mount_page(
            self._selector_or_element,
            self.forbidden_page,
            None,
            {"error": exception},
            handle_exceptions=False,
        )

    def handle_unauthorized(self, exception):
        """
        Handles the exception for unauthorized page. By default, it mounts the self.unauthorized_page class and passes it
        the exception as an argument.

        Args:
            exception (Exception): The exception that occurred.
        """
        self.mount_page(
            self._selector_or_element, self.unauthorized_page, None, {"error": exception}, handle_exceptions=False
        )

    def handle_error(self, exception):
        """
        Handles the exception for application or unknown errors. By default, it mounts the self.error_page class and
        passes it the exception as an argument.

        Args:
            exception (Exception): The exception that occurred.
        """
        self.mount_page(self._selector_or_element, self.error_page, None, {"error": exception}, handle_exceptions=False)
        if is_server_side:
            raise

    def handle_redirect(self, exception):
        """
        Handles a redirect exception by navigating to the given path.

        Args:
            exception (RedirectException): The redirect exception containing the path to navigate to.
        """
        self.router.navigate_to_path(exception.path)



"""
Provides the base classes for PuePy's reactivity system independent of web concerns. These classes are not intended to be used directly, but could be useful for implementing a similar system in a different context.

Classes:
    Listener: A simple class that notifies a collection of callback functions when its `notify` method is called
    ReactiveDict: A dictionary that notifies a listener when it is updated
"""
import logging
from functools import partial


class Listener:
    """
    A simple class that allows you to register callbacks and then notify them all at once.

    Attributes:
        callbacks (list of callables): A list of callback functions to be called when `notify` is called
    """

    def __init__(self):
        self.callbacks = []

    def add_callback(self, callback):
        """
        Adds a callback function to the listener.

        Args:
            callback (callable): The callback function to be added
        """
        self.callbacks.append(callback)

    def remove_callback(self, callback):
        """
        Removes a callback function from the listener.

        Args:
            callback (callable): The callback to be removed
        """
        self.callbacks.remove(callback)

    def notify(self, *args, **kwargs):
        """
        Notify method

        Executes each callback function in the callbacks list by passing in the given arguments and keyword arguments.
        If an exception occurs during the callback execution, it is logged using the logging library.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        for callback in self.callbacks:
            try:
                callback(*args, **kwargs)
            except Exception as e:
                logging.exception("Error in callback for {self}: {callback}:".format(self=self, callback=callback))

    def __str__(self):
        if len(self.callbacks) == 1:
            return f"Listener: {self.callbacks[0]}"
        elif len(self.callbacks) > 1:
            return f"Listener with {len(self.callbacks)} callbacks"
        else:
            return "Listener with no callbacks"

    def __repr__(self):
        return f"<{self}>"


class ReactiveDict(dict):
    """
    A dictionary that notifies a listener when it is updated.

    Attributes:
        listener (Listener): A listener object that is notified when the dictionary is updated
        key_listeners (dict): A dictionary of listeners that are notified when a specific key is updated
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args)
        self.listener = Listener()
        self.key_listeners = {}
        self._in_mutation = False
        self._notifications_pending = set()
        self._keys_mutate = None

    def add_key_listener(self, key, callback):
        """
        Adds a key listener to the object.

        Args:
            key (str): The key for which the listener will be added.
            callback (callable): The callback function to be executed when the key event is triggered.
        """
        if key not in self.key_listeners:
            self.key_listeners[key] = Listener()
        self.key_listeners[key].add_callback(callback)

    def notify(self, *keys):
        """
        Notifies the listener and key listeners that the object has been updated.

        Args:
            *keys: A variable number of keys to be modified for key-specific listeners.
        """
        if keys:
            self._notifications_pending.update(keys)
        else:
            self._notifications_pending.update(self.keys())

        if not self._in_mutation:
            self._flush_pending()

    def mutate(self, *keys):
        """
        To be used as a context manager, this method is for either deferring all notifications until a change has been completed and/or notifying listeners when "deep" changes are made that would have gone undetected by `__setitem__`.

        Examples:
            ``` py
            with reactive_dict.mutate("my_list", "my_dict"):
                reactive_dict["my_list"].append("spam")
                reactive_dict["my_dict"]["spam"] = "eggs"
            ```

        Args:
            *keys: A variable number of keys to update the notifications pending attribute with. If no keys are provided, all keys in the object will be updated.

        Returns:
            The reactive dict itself, which stylistically could be nice to use in a `with` statement.
        """
        if keys:
            self._notifications_pending.update(keys)
        else:
            self._notifications_pending.update(self.keys())
        self._keys_mutate = keys
        return self

    def _flush_pending(self):
        while self._notifications_pending:
            key = self._notifications_pending.pop()
            value = self.get(key, None)
            self.listener.notify(key, value)
            if key in self.key_listeners:
                self.key_listeners[key].notify(key, value)

    def __setitem__(self, key, value):
        if (key in self and value != self[key]) or key not in self:
            super().__setitem__(key, value)
            self.notify(key)

    def __delitem__(self, key):
        super().__delitem__(key)
        self.notify(key)

    def __enter__(self):
        self._in_mutation = True

        if len(self._keys_mutate) == 0:
            return self.get(self._keys_mutate)
        elif len(self._keys_mutate) > 1:
            return [self.get(k) for k in self._keys_mutate]

    def __exit__(self, type, value, traceback):
        self._in_mutation = False
        self._flush_pending()


class Stateful:
    """
    A class that provides a reactive state management system for components. A
    """

    def add_context(self, name: str, value: ReactiveDict):
        """
        Adds contxt from a reactive dict to be reacted on by the component.
        """
        value.listener.add_callback(partial(self._on_state_change, name))

    def initial(self):
        """
        To be overridden in subclasses, the `initial()` method defines the initial state of the stateful object.

        Returns:
            (dict): Initial component state
        """
        return {}

    def on_state_change(self, context, key, value):
        """
        To be overridden in subclasses, this method is called whenever the state of the component changes.

        Args:
            context: What context the state change occured in
            key: The key modified
            value: The new value
        """
        pass

    def _on_state_change(self, context, key, value):
        self.on_state_change(context, key, value)

        if hasattr(self, f"on_{key}_change"):
            getattr(self, f"on_{key}_change")(value)



"""
Common exceptions in the PuePy framework.

Classes:
    ElementNotInDom: Raised when an element is not found in the DOM, but it is expected to be, such as when getting Tag.element
    PropsError: Raised when unexpected props are passed to a component
    PageError: Analogous to http errors, but for a single-page app where the error is client-side
    NotFound: Page not found
    Forbidden: Forbidden
    Unauthorized: Unauthorized
    Redirect: Redirect
"""


class ElementNotInDom(Exception):
    """
    Raised when an element is not found in the DOM, but it is expected to be, such as when getting Tag.element
    """

    pass


class PropsError(ValueError):
    """
    Raised when unexpected props are passed to a component
    """


class PageError(Exception):
    """
    Analogous to http errors, but for a single-page app where the error is client-side
    """

    pass


class NotFound(PageError):
    """
    Raised when the router could not find a page matching the user's URL.
    """

    def __str__(self):
        return "Page not found"


class Forbidden(PageError):
    """
    Raised manually, presumably when the user is not authorized to access a page.
    """

    def __str__(self):
        return "Forbidden"


class Unauthorized(PageError):
    """
    Raised manually, presumably when the user is not authenticated.
    """

    def __str__(self):
        return "Unauthorized"


class Redirect(PageError):
    """
    Raised manually when the user should be redirected to another page.
    """

    def __init__(self, path):
        self.path = path

    def __str__(self):
        return f"Redirect to {self.path}"



import os


def get_version():
    pyproject_path = os.path.join(os.path.dirname(__file__), "..", "pyproject.toml")
    with open(pyproject_path, "r") as pyproject_file:
        for line in pyproject_file.readlines():
            if "version" in line:
                return line.split("=")[1].strip().strip('"')


try:
    __version__ = get_version()
except Exception:
    __version__ = "Unknown; not in a PyProject.toml file."



from .runtime import is_server_side, create_proxy, Object, platform, PLATFORM_PYODIDE

try:
    from pyscript.ffi import to_js
except ImportError:

    def to_js(obj):
        return obj


def mixed_to_underscores(input_string, separator="_"):
    result = []
    for index, char in enumerate(input_string):
        if char.isupper() and index > 0:
            result.append(separator)
        result.append(char.lower())
    return "".join(result)


def merge_classes(*items):
    classes = set()
    exclude_classes = set()

    for class_option in items:
        if isinstance(class_option, str):
            classes.update(class_option.split(" "))
        elif isinstance(class_option, list):
            classes.update(class_option)
        elif isinstance(class_option, dict):
            classes.update([key for key, value in class_option.items() if value])
            exclude_classes.update([key for key, value in class_option.items() if not value])
        elif class_option is None:
            pass
        else:
            classes.update(list(class_option))
    for c in classes:
        if c.startswith("/"):
            exclude_classes.add(c)
            exclude_classes.add(c[1:])
    classes.difference_update(exclude_classes)
    return classes


def jsobj(**kwargs):
    if is_server_side:
        return kwargs
    else:
        return to_js(kwargs)


def _extract_event_handlers(kwargs):
    event_handlers = {}
    for key in list(kwargs.keys()):
        if key.startswith("on_"):
            event_handlers[key[3:]] = kwargs.pop(key)
    return event_handlers


def get_attributes(element):
    if is_server_side:
        return list(element.attributes.keys())
    else:
        return [a.name for a in element.attributes]


def patch_dom_element(source_element, target_element):
    """
    This method patches the target DOM element with attributes and children from the source DOM element. It follows
    the following steps:

    1. Removes attributes from the target element that don't exist in the source element
    2. Sets attributes from the source element to the target element
    3. Iterates over the children of both source and target elements, preserving elements where possible

    :param source_element: The source DOM element that contains the attributes and children to patch.
    :param target_element: The target DOM element that will be patched with the attributes and children from the source
    element.
    """
    # Use morphdom on the client side
    if morphdom:
        return morphdom.default(target_element, source_element)

    # Remove attributes that don't exist in source element
    for attribute in get_attributes(target_element):
        if not source_element.hasAttribute(attribute):
            target_element.removeAttribute(attribute)

    # Set attributes from source to target
    for attribute in get_attributes(source_element):
        target_element.setAttribute(attribute, source_element.getAttribute(attribute))

    if not is_server_side:
        if source_element.tagName.lower() in ("input", "radio", "option", "textarea"):
            target_element.value = source_element.value

    # Iterate over the children
    target_child_nodes = list(target_element.childNodes)
    source_child_nodes = list(source_element.childNodes)

    child_count = max(len(source_child_nodes), len(target_child_nodes))
    for i in range(child_count):
        source_child = source_child_nodes[i] if len(source_child_nodes) > i else None
        target_child = target_child_nodes[i] if len(target_child_nodes) > i else None

        if source_child and target_child:
            if source_child.nodeName == target_child.nodeName and target_child.nodeName == "#text":
                target_child.nodeValue = source_child.nodeValue
            elif source_child.nodeName == target_child.nodeName:
                patch_dom_element(source_child, target_child)
            else:
                target_element.replaceChild(source_child, target_child)

        elif source_child:
            target_element.appendChild(source_child)
        elif target_child:
            target_element.removeChild(target_child)


# Import morphdom if available
morphdom = None
if not is_server_side:
    try:
        from pyscript.js_modules import morphdom
    except ImportError:
        print(
            "Consider loading https://cdn.jsdelivr.net/npm/morphdom@2.7.2/+esm into js_modules for faster"
            " patching and more reliable DOM updates."
        )



"""
Browser Storage Module

This module provides a `BrowserStorage` class that interfaces with browser storage
objects such as `localStorage` and `sessionStorage`. It mimics dictionary-like
behavior for interacting with storage items.


Classes:
    BrowserStorage: A class that provides dictionary-like access to browser storage objects.
"""

try:
    from js import Object

    is_server_side = False
except ImportError:
    is_server_side = True


class BrowserStorage:
    """
    Provides dictionary-like interface to browser storage objects.

    Attributes:
        target: The browser storage object (e.g., localStorage, sessionStorage).
        description (str): Description of the storage instance.

    """

    class NoDefault:
        """Placeholder class for default values when no default is provided."""

        pass

    def __init__(self, target, description):
        """
        Initializes the BrowserStorage instance.

        Args:
            target: The browser storage object.
            description (str): Description of the storage instance.
        """
        self.target = target
        self.description = description

    def __getitem__(self, key):
        """
        Retrieves the value for a given key from the storage.

        Args:
            key (str): The key for the item to retrieve.

        Returns:
            The value associated with the key.

        Raises:
            KeyError: If the key does not exist in the storage.
        """
        value = self.target.getItem(key)
        if value is None:
            raise KeyError(key)
        return value

    def __setitem__(self, key, value):
        """
        Sets the value for a given key in the storage.

        Args:
            key (str): The key for the item to set.
            value: The value to associate with the key.
        """
        self.target.setItem(key, str(value))

    def __delitem__(self, key):
        """
        Deletes the item for a given key from the storage.

        Args:
            key (str): The key for the item to delete.

        Raises:
            KeyError: If the key does not exist in the storage.
        """
        if self.target.getItem(key) is None:
            raise KeyError(key)
        self.target.removeItem(key)

    def __contains__(self, key):
        """
        Checks if a key exists in the storage.

        Args:
            key (str): The key to check.

        Returns:
            bool: True if the key exists, False otherwise.
        """
        return not self.target.getItem(key) is None

    def __len__(self):
        """
        Returns the number of items in the storage.

        Returns:
            int: The number of items in the storage.
        """
        return self.target.length

    def __iter__(self):
        """
        Returns an iterator over the keys in the storage.

        Returns:
            iterator: An iterator over the keys.
        """
        return iter(self.keys())

    def items(self):
        """
        Returns an iterator over the (key, value) pairs in the storage.

        Yields:
            tuple: (key, value) pairs in the storage.
        """
        for item in Object.entries(self.target):
            yield item[0], item[1]

    def keys(self):
        """
        Returns a list of keys in the storage.

        Returns:
            list: A list of keys.
        """
        return list(Object.keys(self.target))

    def get(self, key, default=None):
        """
        Retrieves the value for a given key, returning a default value if the key does not exist.

        Args:
            key (str): The key for the item to retrieve.
            default: The default value to return if the key does not exist.

        Returns:
            The value associated with the key, or the default value.
        """
        value = self.target.getItem(key)
        if value is None:
            return default
        else:
            return value

    def clear(self):
        """
        Clears all items from the storage.
        """
        self.target.clear()

    def copy(self):
        """
        Returns a copy of the storage as a dictionary.

        Returns:
            dict: A dictionary containing all items in the storage.
        """
        return dict(self.items())

    def pop(self, key, default=NoDefault):
        """
        Removes the item with the given key from the storage and returns its value.

        Args:
            key (str): The key for the item to remove.
            default: The default value to return if the key does not exist.

        Returns:
            The value associated with the key, or the default value.

        Raises:
            KeyError: If the key does not exist and no default value is provided.
        """
        value = self.target.getItem(key)
        if value is None and default is self.NoDefault:
            raise KeyError(key)
        else:
            self.target.removeItem(key)
            return value

    def popitem(self):
        """
        Not implemented. Raises NotImplementedError.

        Raises:
            NotImplementedError: Always raised as the method is not implemented.
        """
        raise NotImplementedError("popitem not implemented")

    def reversed(self):
        """
        Not implemented. Raises NotImplementedError.

        Raises:
            NotImplementedError: Always raised as the method is not implemented.
        """
        raise NotImplementedError("reversed not implemented")

    def setdefault(self, key, default=None):
        """
        Sets the value for the key if it does not already exist in the storage.

        Args:
            key (str): The key for the item.
            default: The value to set if the key does not exist.

        Returns:
            The value associated with the key, or the default value.
        """
        if key in self:
            return self[key]
        else:
            self[key] = default
            return default

    def update(self, other):
        """
        Updates the storage with items from another dictionary or iterable of key-value pairs.

        Args:
            other: A dictionary or iterable of key-value pairs to update the storage with.
        """
        for k, v in other.items():
            self[k] = v

    def values(self):
        """
        Returns a list of values in the storage.

        Returns:
            list: A list of values.
        """
        return list(Object.values(self.target))

    def __str__(self):
        return self.description

    def __repr__(self):
        return f"<{self}>"



import sys

PLATFORM_PYODIDE = "pyodide"
PLATFORM_MICROPYTHON = "micropython"
PLATFORM_CPYTHON = "cpython"


def _detect_platform():
    if sys.platform == "emscripten":
        return PLATFORM_PYODIDE
    elif sys.platform == "webassembly" and sys.implementation.name == "micropython":
        return PLATFORM_MICROPYTHON
    elif sys.implementation.name == "cpython":
        return PLATFORM_CPYTHON


platform = _detect_platform()
is_server_side = platform not in (PLATFORM_PYODIDE, PLATFORM_MICROPYTHON)

if platform == PLATFORM_PYODIDE:
    from pyodide.ffi import create_proxy
    from pyodide.ffi.wrappers import add_event_listener, remove_event_listener
elif platform == PLATFORM_MICROPYTHON:
    from pyscript.ffi import create_proxy

    from js import addEventListener, removeEventListener

    def add_event_listener(elt, event, listener):
        return elt.addEventListener(event, listener)

    def remove_event_listener(elt, event, listener):
        return elt.removeEventListener(event, create_proxy(listener))

    # add_event_listener, remove_event_listener = addEventListener, removeEventListener

if is_server_side:
    document = setTimeout = Object = CustomEvent = window = history = add_event_listener = remove_event_listener = None

    def create_proxy(obj):
        return obj

else:
    from js import document, setTimeout, Object, CustomEvent, window, history



import binascii
import hashlib

from .exceptions import ElementNotInDom, PropsError, PageError
from .reactivity import ReactiveDict, Stateful
from .runtime import (
    add_event_listener,
    remove_event_listener,
    create_proxy,
    document,
    is_server_side,
    setTimeout,
    CustomEvent,
)
from .util import (
    mixed_to_underscores,
    merge_classes,
    _extract_event_handlers,
    patch_dom_element,
)


class Prop:
    """
    Class representing a prop for a component.

    Attributes:
        name (str): The name of the property.
        description (str): The description of the property (optional).
        type (type): The data type of the property (default: str).
        default_value: The default value of the property (optional).
    """

    def __init__(self, name, description=None, type=str, default_value=None):
        self.name = name
        self.description = description
        self.type = type
        self.default_value = default_value


def _element_input_type(element):
    try:
        if element.tagName.lower() == "input" and element.getAttribute("type"):
            return element.getAttribute("type").lower()
    except:
        return None


class Tag:
    """
    The most basic building block of a PuePy app. A Tag is a single HTML element. This is also the base class of
    `Component`, which is then the base class of `Page`.

    Attributes:
        default_classes (list): Default classes for the tag.
        default_attrs (dict): Default attributes for the tag.
        default_role (str): Default role for the tag.
        page (Page): The page the tag is on.
        router (Router or None): The router the application is using, if any.
        parent (Tag): The parent tag, component, or page.
        application (Application): The application instance.
        element: The rendered element on the DOM. Raises ElementNotInDom if not found.
        children (list): The children of the tag.
        refs (dict): The refs of the tag.
        tag_name (str): The name of the tag.
        ref (str): The reference of the tag.
    """

    stack = []
    population_stack = []
    origin_stack = [[]]
    component_stack = []
    default_classes = []
    default_attrs = {}
    default_role = None

    document = document

    def __init__(
        self,
        tag_name,
        ref,
        page: "Page" = None,
        parent=None,
        parent_component=None,
        origin=None,
        children=None,
        **kwargs,
    ):
        # Kept so we can garbage collect them later
        self._added_event_listeners = []

        # Ones manually added, which we persist when reconfigured
        self._manually_added_event_listeners = {}

        # The rendered element
        self._rendered_element = None

        # Child nodes and origin refs
        self.children = []
        self.refs = {}

        self.tag_name = tag_name
        self.ref = ref

        # Attrs that webcomponents create that we need to preserve
        self._retained_attrs = {}

        # Add any children passed to constructor
        if children:
            self.add(*children)

        # Configure self._page
        if isinstance(page, Page):
            self._page = page
        elif isinstance(self, Page):
            self._page = self
        elif page:
            raise Exception(f"Unknown page type {type(page)}")
        else:
            raise Exception("No page passed")

        if id in kwargs:
            self._element_id = kwargs["id"]
        elif self._page and self._page.application:
            self._element_id = self._page.application.element_id_generator.get_id_for_element(self)
        else:
            self._element_id = f"ppauto-{id(self)}"

        if isinstance(parent, Tag):
            self.parent = parent
            parent.add(self)
        elif parent:
            raise Exception(f"Unknown parent type {type(parent)}: {repr(parent)}")
        else:
            self.parent = None

        if isinstance(parent_component, Component):
            self.parent_component = parent_component
        elif parent_component:
            raise Exception(f"Unknown parent_component type {type(parent_component)}: {repr(parent_component)}")
        else:
            self.parent_component = None

        self.origin = origin
        self._children_generated = False

        self._configure(kwargs)

    def __del__(self):
        if not is_server_side:
            while self._added_event_listeners:
                remove_event_listener(*self._added_event_listeners.pop())

    @property
    def application(self):
        return self._page._application

    def _configure(self, kwargs):
        self._kwarg_event_listeners = _extract_event_handlers(kwargs)
        self._handle_bind(kwargs)
        self._handle_attrs(kwargs)

    def _handle_bind(self, kwargs):
        if "bind" in kwargs:
            self.bind = kwargs.pop("bind")
            if "value" in kwargs:
                raise Exception("Cannot specify both 'bind' and 'value'")
        else:
            self.bind = None

    def _handle_attrs(self, kwargs):
        self.attrs = self._retained_attrs.copy()
        for k, v in kwargs.items():
            if hasattr(self, f"set_{k}"):
                getattr(self, f"set_{k}")(v)
            else:
                self.attrs[k] = v

    def populate(self):
        """To be overwritten by subclasses, this method will define the composition of the element"""
        pass

    def precheck(self):
        """
        Before doing too much work, decide whether rendering this tag should raise an error or not. This may be useful,
        especially on a Page, to check if the user is authorized to view the page, for example:

        Examples:
            ``` py
            def precheck(self):
                if not self.application.state["authenticated_user"]:
                    raise exceptions.Unauthorized()
            ```
        """
        pass

    def generate_children(self):
        """
        Runs populate, but first adds self to self.population_stack, and removes it after populate runs.

        That way, as populate is executed, self.population_stack can be used to figure out what the innermost populate()
        method is being run and thus, where to send bind= parameters.
        """
        self.origin_stack.append([])
        self._refs_pending_removal = self.refs.copy()
        self.refs = {}
        self.population_stack.append(self)
        try:
            self.precheck()
            self.populate()
        finally:
            self.population_stack.pop()
            self.origin_stack.pop()

    def render(self):
        attrs = self.get_default_attrs()
        attrs.update(self.attrs)

        element = self._create_element(attrs)

        self._render_onto(element, attrs)
        self.post_render(element)
        return element

    def _create_element(self, attrs):
        if "xmlns" in attrs:
            element = self.document.createElementNS(attrs.get("xmlns"), self.tag_name)
        else:
            element = self.document.createElement(self.tag_name)

        element.setAttribute("id", self.element_id)
        if is_server_side:
            element.setIdAttribute("id")

        self.configure_element(element)

        return element

    def configure_element(self, element):
        pass

    def post_render(self, element):
        pass

    @property
    def element_id(self):
        return self._element_id

    @property
    def element(self):
        el = self.document.getElementById(self.element_id)
        if el:
            return el
        else:
            raise ElementNotInDom(self.element_id)

    def _render_onto(self, element, attrs):
        self._rendered_element = element

        # Handle classes
        classes = self.get_render_classes(attrs)

        if classes:
            # element.className = " ".join(classes)
            element.setAttribute("class", " ".join(classes))

        # Add attributes
        for key, value in attrs.items():
            if key not in ("class_name", "classes", "class"):
                if hasattr(self, f"handle_{key}_attr"):
                    getattr(self, f"handle_{key}_attr")(element, value)
                else:
                    if key.endswith("_"):
                        attr = key[:-1]
                    else:
                        attr = key
                    attr = attr.replace("_", "-")

                    if isinstance(value, bool) or value is None:
                        if value:
                            element.setAttribute(attr, attr)
                    elif isinstance(value, (str, int, float)):
                        element.setAttribute(attr, value)
                    else:
                        element.setAttribute(attr, str(value))

        if "role" not in attrs and self.default_role:
            element.setAttribute("role", self.default_role)

        # Add event handlers
        self._add_listeners(element, self._kwarg_event_listeners)
        self._add_listeners(element, self._manually_added_event_listeners)

        # Add bind
        if self.bind and self.origin:
            if _element_input_type(element) == "checkbox":
                if is_server_side and self.origin.state[self.bind]:
                    element.setAttribute("checked", self.origin.state[self.bind])
                else:
                    element.checked = bool(self.origin.state[self.bind])
            else:
                if is_server_side:
                    element.setAttribute("value", self.origin.state[self.bind])
                else:
                    element.value = self.origin.state[self.bind]

            self._add_event_listener(element, "input", self.on_bind_input)
        elif self.bind:
            raise Exception("Cannot specify bind a valid parent component")

        self.render_children(element)

    def _add_listeners(self, element, listeners):
        for key, value in listeners.items():
            key = key.replace("_", "-")
            if isinstance(value, (list, tuple)):
                for handler in value:
                    self._add_event_listener(element, key, handler)
            else:
                self._add_event_listener(element, key, value)

    def render_children(self, element):
        for child in self.children:
            if isinstance(child, Slot):
                if child.children:  # If slots don't have any children, don't bother.
                    element.appendChild(child.render())
            elif isinstance(child, Tag):
                element.appendChild(child.render())
            elif isinstance(child, html):
                element.insertAdjacentHTML("beforeend", str(child))
            elif isinstance(child, str):
                element.appendChild(self.document.createTextNode(child))
            elif child is None:
                pass
            elif getattr(child, "nodeType", None) is not None:
                # DOM element
                element.appendChild(child)
            else:
                self.render_unknown_child(element, child)

    def render_unknown_child(self, element, child):
        """
        Called when the child is not a Tag, Slot, or html. By default, it raises an error.
        """
        raise Exception(f"Unknown child type {type(child)} onto {self}")

    def get_render_classes(self, attrs):
        return merge_classes(
            set(self.get_default_classes()),
            attrs.pop("class_name", []),
            attrs.pop("classes", []),
            attrs.pop("class", []),
        )

    def get_default_classes(self):
        """
        Returns a shallow copy of the default_classes list.

        This could be overridden by subclasses to provide a different default_classes list.

        Returns:
            (list): A shallow copy of the default_classes list.
        """
        return self.default_classes.copy()

    def get_default_attrs(self):
        return self.default_attrs.copy()

    def _add_event_listener(self, element, event, listener):
        """
        Just an internal wrapper around add_event_listener (JS function) that keeps track of what we added, so
        we can garbage collect it later.

        Should probably not be used outside this class.
        """
        self._added_event_listeners.append((element, event, listener))
        if not is_server_side:
            add_event_listener(element, event, listener)

    def add_event_listener(self, event, handler):
        """
        Add an event listener for a given event.

        Args:
            event (str): The name of the event to listen for.
            handler (function): The function to be executed when the event occurs.

        """
        if event not in self._manually_added_event_listeners:
            self._manually_added_event_listeners[event] = handler
        else:
            existing_handler = self._manually_added_event_listeners[event]
            if isinstance(existing_handler, (list, tuple)):
                self._manually_added_event_listeners[event] = [existing_handler] + list(handler)
            else:
                self._manually_added_event_listeners[event] = [existing_handler, handler]
        if self._rendered_element:
            self._add_event_listener(self._rendered_element, event, handler)

    def mount(self, selector_or_element):
        self.update_title()
        if not self._children_generated:
            with self:
                self.generate_children()

        if isinstance(selector_or_element, str):
            element = self.document.querySelector(selector_or_element)
        else:
            element = selector_or_element

        if not element:
            raise RuntimeError(f"Element {selector_or_element} not found")

        element.innerHTML = ""
        element.appendChild(self.render())
        self.recursive_call("on_ready")

    def recursive_call(self, method, *args, **kwargs):
        """
        Recursively call a specified method on all child Tag objects.

        Args:
            method (str): The name of the method to be called on each Tag object.
            *args: Optional arguments to be passed to the method.
            **kwargs: Optional keyword arguments to be passed to the method.
        """
        for child in self.children:
            if isinstance(child, Tag):
                child.recursive_call(method, *args, **kwargs)
        getattr(self, method)(*args, **kwargs)

    def on_ready(self):
        pass

    def _retain_implicit_attrs(self):
        """
        Some web components add attributes after their elements are rendered. Eg, <sl-button> becomes <sl-button variant="default">

        If we patch the DOM and remove those attributes, we lose them. This method is called recursively on redraw
        to retain any attributes that were added by the web component.
        """
        if getattr(self.element, "shadowRoot"):
            for attr in self.element.attributes:
                if attr.name not in self.attrs and attr.name != "id":
                    self._retained_attrs[attr.name] = attr.value

    def on_redraw(self):
        pass

    def on_bind_input(self, event):
        if _element_input_type(event.target) == "checkbox":
            self.origin.state[self.bind] = event.target.checked
        elif _element_input_type(event.target) == "number":
            value = event.target.value
            try:
                if "." in str(value):
                    value = float(value)
                else:
                    value = int(value)
            except (ValueError, TypeError):
                pass
            self.origin.state[self.bind] = value
        else:
            self.origin.state[self.bind] = event.target.value

    @property
    def page(self):
        if self._page:
            return self._page
        elif isinstance(self, Page):
            return self

    @property
    def router(self):
        if self.application:
            return self.application.router

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, new_parent):
        existing_parent = getattr(self, "_parent", None)
        if new_parent == existing_parent:
            if new_parent and self not in new_parent.children:
                existing_parent.children.append(self)
            return

        if existing_parent and self in existing_parent.children:
            existing_parent.children.remove(self)
        if new_parent and self not in new_parent.children:
            new_parent.children.append(self)

        self._parent = new_parent

    def add(self, *children):
        for child in children:
            if isinstance(child, Tag):
                child.parent = self
            else:
                self.children.append(child)

    def redraw(self):
        if self in self.page.redraw_list:
            self.page.redraw_list.remove(self)

        try:
            element = self.element
        except ElementNotInDom:
            return

        if is_server_side:
            old_active_element_id = None
        else:
            old_active_element_id = self.document.activeElement.id if self.document.activeElement else None

            self.recursive_call("_retain_implicit_attrs")

        self.children = []

        attrs = self.get_default_attrs()
        attrs.update(self.attrs)

        self.update_title()
        with self:
            self.generate_children()

        staging_element = self._create_element(attrs)

        self._render_onto(staging_element, attrs)

        patch_dom_element(staging_element, element)

        if old_active_element_id is not None:
            el = self.document.getElementById(old_active_element_id)
            if el:
                el.focus()

        self.recursive_call("on_redraw")

    def trigger_event(self, event, detail=None, **kwargs):
        """
                Triggers an event to be consumed by code using this class.

                Args:
                    event (str): The name of the event to trigger. If the event name contains underscores, a warning message is printed suggesting to use dashes instead.
                    detail (dict, optional): Additional data to be sent with the event. This should be a dictionary where the keys and values will be converted to JavaScript objects.
                    **kwargs: Additional keyword arguments. These arguments are not used in the implementation of the method and are ignored.
        ß"""
        if "_" in event:
            print("Triggering event with underscores. Did you mean dashes?: ", event)

        # noinspection PyUnresolvedReferences
        from pyscript.ffi import to_js

        # noinspection PyUnresolvedReferences
        from js import Object, Map

        if detail:
            event_object = to_js({"detail": Map.new(Object.entries(to_js(detail)))})
        else:
            event_object = to_js({})

        self.element.dispatchEvent(CustomEvent.new(event, event_object))

    def update_title(self):
        """
        To be overridden by subclasses (usually pages), this method should update the Window title as needed.

        Called on mounting or redraw.
        """
        pass

    def __enter__(self):
        self.stack.append(self)
        self.origin_stack[0].append(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stack.pop()
        self.origin_stack[0].pop()
        return False

    def __str__(self):
        return self.tag_name

    def __repr__(self):
        return f"<{self} ({id(self)})>"


class Slot(Tag):
    def __init__(self, ref, slot_name="default", tag_name="span", **kwargs):
        self.slot_name = slot_name
        super().__init__(tag_name=tag_name, ref=ref, **kwargs)

    def __str__(self):
        return f"Slot: {self.slot_name}"


class Component(Tag, Stateful):
    """
    Components are a way of defining reusable and composable elements in PuePy. They are a subclass of Tag, but provide
    additional features such as state management and props. By defining your own components and registering them, you
    can create a library of reusable elements for your application.

    Attributes:
        enclosing_tag (str): The tag name that will enclose the component. To be defined as a class attribute on subclasses.
        component_name (str): The name of the component. If left blank, class name is used. To be defined as a class attribute on subclasses.
        redraw_on_state_changes (bool): Whether the component should redraw when its state changes. To be defined as a class attribute on subclasses.
        redraw_on_app_state_changes (bool): Whether the component should redraw when the application state changes. To be defined as a class attribute on subclasses.
        props (list): A list of props for the component. To be defined as a class attribute on subclasses.
    """

    enclosing_tag = "div"
    component_name = None
    redraw_on_state_changes = True
    redraw_on_app_state_changes = True

    props = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, tag_name=self.enclosing_tag, **kwargs)
        self.state = ReactiveDict(self.initial())
        self.add_context("state", self.state)

        self.slots = {}

    def _handle_attrs(self, kwargs):
        self._handle_props(kwargs)

        super()._handle_attrs(kwargs)

    def _handle_props(self, kwargs):
        if not hasattr(self, "props_expanded"):
            self._expanded_props()

        self.props_values = {}
        for name, prop in self.props_expanded.items():
            value = kwargs.pop(prop.name, prop.default_value)
            setattr(self, name, value)
            self.props_values[name] = value

    @classmethod
    def _expanded_props(cls):
        # This would be ideal for metaprogramming, but we do it this way to be compatible with Micropython. :/
        props_expanded = {}
        for prop in cls.props:
            if isinstance(prop, Prop):
                props_expanded[prop.name] = prop
            elif isinstance(prop, dict):
                props_expanded[prop["name"]] = Prop(**prop)
            elif isinstance(prop, str):
                props_expanded[prop] = Prop(name=prop)
            else:
                raise PropsError(f"Unknown prop type {type(prop)}")
        cls.props_expanded = props_expanded

    def initial(self):
        """
        To be overridden in subclasses, the `initial()` method defines the initial state of the component.

        Returns:
            (dict): Initial component state
        """
        return {}

    def _on_state_change(self, context, key, value):
        super()._on_state_change(context, key, value)

        if context == "state":
            redraw_rule = self.redraw_on_state_changes
        elif context == "app":
            redraw_rule = self.redraw_on_app_state_changes
        else:
            return

        if redraw_rule is True:
            self.page.redraw_tag(self)
        elif redraw_rule is False:
            pass
        elif isinstance(redraw_rule, (list, set)):
            if key in redraw_rule:
                self.page.redraw_tag(self)
        else:
            raise Exception(f"Unknown value for redraw rule: {redraw_rule} (context: {context})")

    def insert_slot(self, name="default", **kwargs):
        """
        In defining your own component, when you want to create a slot in your `populate` method, you can use this method.

        Args:
            name (str): The name of the slot. If not passed, the default slot is inserted.
            **kwargs: Additional keyword arguments to be passed to Slot initialization.

        Returns:
            Slot: The inserted slot object.
        """
        if name in self.slots:
            self.slots[name].parent = Tag.stack[-1]  # The children will be cleared during redraw, so re-establish
        else:
            self.slots[name] = Slot(ref=f"slot={name}", slot_name=name, page=self.page, parent=Tag.stack[-1], **kwargs)
        slot = self.slots[name]
        if self.origin:
            slot.origin = self.origin
            if slot.ref:
                self.origin.refs[slot.ref] = slot
        return slot

    def slot(self, name="default"):
        """
        To be used in the `populate` method of code making use of this component, this method returns the slot object
        with the given name. It should be used inside of a context manager.

        Args:
            name (str): The name of the slot to clear and return.

        Returns:
            Slot: The cleared slot object.
        """
        #
        # We put this here, so it clears the children only when the slot-filler is doing its filling.
        # Otherwise, the previous children are kept. Lucky them.
        self.slots[name].children = []
        return self.slots[name]

    def __enter__(self):
        self.stack.append(self)
        self.origin_stack[0].append(self)
        self.component_stack.append(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stack.pop()
        self.origin_stack[0].pop()
        self.component_stack.pop()
        return False

    def __str__(self):
        return f"{self.component_name or self.__class__.__name__} ({self.ref} {id(self)})"

    def __repr__(self):
        return f"<{self}>"


class Page(Component):
    def __init__(self, matched_route=None, application=None, **kwargs):
        ref = mixed_to_underscores(self.__class__.__name__)
        self.matched_route = matched_route
        self._application = application

        self._redraw_timeout_set = False
        self.redraw_list = set()

        super().__init__(ref=ref, **kwargs)
        if self.application:
            self.add_context("app", self.application.state)

    def update_title(self):
        title = self.page_title()
        if title is not None:
            self.document.title = title

    def page_title(self):
        pass

    def redraw_tag(self, tag):
        assert isinstance(tag, Tag)
        self.redraw_list.add(tag)

        if not self._redraw_timeout_set:
            if is_server_side:
                self._do_redraw()
            else:
                setTimeout(create_proxy(self._do_redraw), 1)
                self._redraw_timeout_set = True

    def _do_redraw(self):
        try:
            redrawn_already = set()
            try:
                while self.redraw_list:
                    tag = self.redraw_list.pop()
                    if tag in redrawn_already:
                        raise Exception(
                            "A redraw event is causing a circular redraw event loop. Yo dawg, are you causing "
                            "a redraw from your redraw?"
                        )

                    tag.redraw()
                    redrawn_already.add(tag)
            finally:
                self._redraw_timeout_set = False
        except PageError as e:
            self.application.handle_page_error(e)
        except Exception as e:
            self.application.handle_error(e)


class Builder:
    def __init__(self):
        self.components = {}

    def generate_tag(self, tag_name, *children, **kwargs):
        tag_name = tag_name.lower().strip()
        if kwargs.get("tag"):
            tag_name = kwargs.pop("tag")
        elif "_" in tag_name:
            tag_name = tag_name.replace("_", "-")

        if tag_name == "insert_slot":
            print(f"Called t.insert_slot. Did you mean self.insert_slot?")
        elif tag_name == "slot":
            print(f"Called t.slot. Did you mean <component>.slot?")

        parent = Tag.stack[-1] if Tag.stack else None
        parent_component = Tag.component_stack[-1] if Tag.component_stack else None
        root_tag = Tag.stack[0] if Tag.stack else None
        origin = Tag.population_stack[-1] if Tag.population_stack else None

        # The root tag might not always be the page, but if it isn't, it should have a reference to the
        # actual page.
        if isinstance(root_tag, Page):
            page = root_tag
        elif root_tag:
            page = root_tag.page
        else:
            raise Exception("t.generate_tag called without a context")

        # Determine ref value
        ref_part = f"__{parent.ref}.{tag_name}{len(parent.children) + 1}_" + binascii.hexlify(
            hashlib.sha256(str(kwargs).encode()).digest()
        )[:8].decode("utf8")

        ref = kwargs.pop("ref", ref_part)

        if ref and origin and ref in origin._refs_pending_removal:
            element: Tag = origin._refs_pending_removal.pop(ref)

            assert element.origin == origin
            element._configure(kwargs)
            element.children = []
            if children:
                element.add(*children)
            element.parent = parent
        elif tag_name in self.components:
            element = self.components[tag_name](
                ref=ref,
                page=page,
                parent=parent,
                parent_component=parent_component,
                origin=origin,
                children=children,
                **kwargs,
            )
        else:
            element = Tag(
                tag_name,
                ref=ref,
                page=page,
                parent=parent,
                parent_component=parent_component,
                origin=origin,
                children=children,
                **kwargs,
            )
        origin.refs[ref] = element
        with element:
            element.generate_children()
        return element

    def add_library(self, library):
        for component in library.components:
            self.add_component(component)

    def component(self):
        def inner(cls):
            self.add_component(cls)
            return cls

        return inner

    def add_component(self, component: Component):
        if component.component_name:
            component_name = component.component_name.replace("_", "-")
        else:
            component_name = mixed_to_underscores(component.__name__, "-")
        self.components[component_name.lower()] = component

    def __call__(self, *texts):
        parent = Tag.stack[-1] if Tag.stack else None
        if not parent:
            raise Exception("Cannot call t() to generate text without a parent")
        parent.add(*texts)

    def __getattr__(self, name):
        def _tag(*children, **kwargs):
            return self.generate_tag(name, *children, **kwargs)

        return _tag


t = Builder()


class html(str):
    pass
